/********************************************************************************************
 *                                                                                          *
 * Copyright (C) 2017 Armin Felder, Dennis Beier                                            *
 * This file is part of RocketChatMobileEngine <https://git.fairkom.net/chat/fairchat>.     *
 *                                                                                          *
 * RocketChatMobileEngine is free software: you can redistribute it and/or modify           *
 * it under the terms of the GNU General Public License as published by                     *
 * the Free Software Foundation, either version 3 of the License, or                        *
 * (at your option) any later version.                                                      *
 *                                                                                          *
 * RocketChatMobileEngine is distributed in the hope that it will be useful,                *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of                           *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                            *
 * GNU General Public License for more details.                                             *
 *                                                                                          *
 * You should have received a copy of the GNU General Public License                        *
 * along with RocketChatMobileEngine. If not, see <http://www.gnu.org/licenses/>.           *
 *                                                                                          *
 ********************************************************************************************/


#include "rocketchatchannel.h"
#include "segfaulthandler.h"



RocketChatChannel::RocketChatChannel( RocketChatServerData *pServer, MessageService *pMessageService, QString pRoomId, QString pName, QString pType )
{
    Q_UNUSED( pServer )
    Q_UNUSED( pMessageService )
    this->mRoomId = pRoomId;
    this->mName = pName;
    this->mType = pType;
}

RocketChatChannel::~RocketChatChannel()
{
    mUsers.clear();
}

bool RocketChatChannel::addMessage( const ChatMessage pMessage, bool nosignal )
{
    if ( !pMessage.isNull() ) {
        if ( mMessages.add( pMessage->getId(), pMessage ) ) {
            qint64 timestamp = pMessage.data()->getTimestamp();
            qint64 youngest = mMessages.youngest().data()->getTimestamp();

            if ( timestamp >= youngest && !nosignal ) {
                emit( messageAdded( getRoomId(), timestamp ) );
            }

            return true;
        }
    }

    return false;
}

QHash<QString, QSharedPointer< RocketChatMessage >> RocketChatChannel::getMessages() const
{
    return mMessages.getElements();
}

QString RocketChatChannel::getRoomId() const
{
    return mRoomId;
}

void RocketChatChannel::setRoomId( const QString &pValue )
{
    mRoomId = pValue;
}

QString RocketChatChannel::getName() const
{
    return mName;
}

void RocketChatChannel::setName( const QString &value )
{
    mName = value;
}


bool RocketChatChannel::addUser( QSharedPointer<RocketChatUser> user )
{
    QString username = user->getUserName();

    if ( mUsers.empty() || !mUsers.contains( username ) ) {
        mUsers[username] = user;
        return true;
    }

    return false;
}

QMap<QString, QSharedPointer<RocketChatUser> > RocketChatChannel::getUsers() const
{
    return mUsers;
}

void RocketChatChannel::setUsers( const QMap<QString, QSharedPointer<RocketChatUser> > &value )
{
    mUsers = value;
}

void RocketChatChannel::setUsers( const QList<QSharedPointer<RocketChatUser> > &value )
{
    mUsers.clear();

    for ( const auto &currentUser : value ) {
        mUsers[currentUser->getUserName()] = currentUser;
    }

}

ChatMessage RocketChatChannel::getYoungestMessage()
{
    if ( !mMessages.isEmpty() ) {
        return mMessages.youngest();
    } else {
        return ChatMessage( nullptr );
    }
}

ChatMessage RocketChatChannel::getOldestMessage()
{
    if ( !mMessages.isEmpty() ) {
        return mMessages.oldest();
    } else {
        return ChatMessage( nullptr );
    }
}

MessageRepository *RocketChatChannel::getMessageRepo()
{
    return &mMessages;
}

QList<QSharedPointer<RocketChatMessage>> RocketChatChannel::addMessages( QList<QSharedPointer<RocketChatMessage>> messagesList )
{
    QList<QSharedPointer<RocketChatMessage>> newMessages;

    for ( const auto &currentMessage : messagesList ) {
        if ( !currentMessage.isNull() ) {
            if ( addMessage( currentMessage ) ) {
                newMessages.append( currentMessage );
            }
        }
    }

    emit messageAdded( getRoomId(), 0 );
    return newMessages;

}

bool RocketChatChannel::getJoined() const
{
    return mJoined;
}

void RocketChatChannel::setJoined( bool value )
{
    mJoined = value;
}

unsigned int RocketChatChannel::getUnreadMessages() const
{
    return mUnreadMessages;
}

void RocketChatChannel::setUnreadMessages( unsigned int value )
{
    qDebug() << "set unread to: " << value;
    mUnreadMessages = value;
    emit unreadMessagesChanged( mRoomId, value );
}

QString RocketChatChannel::getType() const
{
    return mType;
}


bool RocketChatChannel::getReadOnly() const
{
    return mReadOnly;
}

void RocketChatChannel::setReadOnly( bool value )
{
    mReadOnly = value;
}

QStringList RocketChatChannel::getMuted() const
{
    return mMuted;
}

void RocketChatChannel::setMuted( const QStringList &value )
{
    mMuted = value;
}

bool RocketChatChannel::getArchived() const
{
    return mArchived;
}

void RocketChatChannel::setArchived( bool value )
{
    mArchived = value;
}


QJsonObject RocketChatChannel::toJsonObject()
{
    QJsonObject currentChannelObject;
    currentChannelObject[QStringLiteral( "name" )] =       getName();
    currentChannelObject[QStringLiteral( "joined" )] =     getJoined();
    currentChannelObject[QStringLiteral( "ro" )] =         getReadOnly();
    currentChannelObject[QStringLiteral( "muted" )] =      QJsonArray::fromStringList( getMuted() );
    currentChannelObject[QStringLiteral( "archived" )] =   getArchived();
    currentChannelObject[QStringLiteral( "type" )] =       getType();
    currentChannelObject[QStringLiteral( "roomId" )] =     getRoomId();
    return currentChannelObject;
}

bool RocketChatChannel::getSelfMuted() const
{
    return mSelfMuted;
}

void RocketChatChannel::setSelfMuted( bool value )
{
    mSelfMuted = value;
}

int RocketChatChannel::operator >( RocketChatChannel &channel )
{
    auto youngestMessage = this->getYoungestMessage();
    auto youngestMessage2 = channel.getYoungestMessage();
    int timestamp1 = youngestMessage.isNull() ? 0 : youngestMessage->getTimestamp() ;
    int timestamp2 = youngestMessage2.isNull() ? 0 : youngestMessage2->getTimestamp() ;
    return timestamp1 > timestamp2;
}

bool RocketChatChannel::getDeleted() const
{
    return mDeleted;
}

void RocketChatChannel::setDeleted( bool deleted )
{
    mDeleted = deleted;
    emit channelDeleted( this->getRoomId(), deleted );
}
bool RocketChatChannel::getBlocked() const
{
    return mBlocked;
}

void RocketChatChannel::setBlocked( bool blocked )
{
    mBlocked = blocked;
    emit dataChanged( mRoomId, "blocked" );
}
