/********************************************************************************************
 *                                                                                          *
 * Copyright (C) 2017 Armin Felder, Dennis Beier                                            *
 * This file is part of RocketChatMobileEngine <https://git.fairkom.net/chat/fairchat>.     *
 *                                                                                          *
 * RocketChatMobileEngine is free software: you can redistribute it and/or modify           *
 * it under the terms of the GNU General Public License as published by                     *
 * the Free Software Foundation, either version 3 of the License, or                        *
 * (at your option) any later version.                                                      *
 *                                                                                          *
 * RocketChatMobileEngine is distributed in the hope that it will be useful,                *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of                           *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                            *
 * GNU General Public License for more details.                                             *
 *                                                                                          *
 * You should have received a copy of the GNU General Public License                        *
 * along with RocketChatMobileEngine. If not, see <http://www.gnu.org/licenses/>.           *
 *                                                                                          *
 ********************************************************************************************/



#include "messageservice.h"



//MessageService::MessageService( PersistanceLayer *pPersistanceLayer,
//                                RocketChatServerData *pServer,
//                                QHash<QString, QString> &pEmojisMap ): mEmojisMap( pEmojisMap )
//{
//    this->mPersistanceLayer = pPersistanceLayer;
//    this->mServer = pServer;
//}

MessageService::MessageService( PersistanceLayer *pPersistanceLayer,
                                RocketChatServerData *pServer,
                                EmojiRepo *pEmojiRepo,
                                MessagesModel *pMessagesModel ): mEmojiRepo( pEmojiRepo ), mMessagesModel( pMessagesModel )
{
    this->mPersistanceLayer = pPersistanceLayer;
    this->mServer = pServer;
}

void MessageService::loadHistory( QSharedPointer<LoadHistoryServiceRequest> pRequest )
{
    if ( !pRequest.isNull() ) {
        switch ( pRequest->getSource() ) {
            case LoadHistoryServiceRequest::Source::DataBase:
                loadHistoryFromDb( pRequest );
                break;

            case LoadHistoryServiceRequest::Source::SERVER:
                loadHistoryFromServer( pRequest );
                break;

            default:
                loadHistoryFromAutoSource( pRequest );
                break;
        }
    }
}



void MessageService::checkForMissingMessages( QSharedPointer<LoadMissedMessageServiceRequest> pRequest )
{
    Q_UNUSED( pRequest )
    
}

void MessageService::sendMessage( QSharedPointer<RocketChatMessage> pMessage )
{
    if ( !pMessage.isNull() ) {
        QJsonObject params;
        params[QStringLiteral( "rid" )] = pMessage->getRoomId();
        params[QStringLiteral( "msg" )] = pMessage->getMessageString();
        QSharedPointer<DDPRequest> sendMessageRequest( new DDPMethodRequest( QStringLiteral( "sendMessage" ), {params} ) );
        mServer->sendDdprequest( sendMessageRequest, true );
    }
}

void MessageService::persistMessage( QSharedPointer<RocketChatMessage> pMessage )
{
    if ( !pMessage.isNull() ) {
        QJsonDocument doc( pMessage->getData() );
        QString msg = doc.toJson();
        mPersistanceLayer->addMessage( pMessage->getId(), pMessage->getRoomId(), pMessage->getAuthor(), pMessage->getTimestamp(), msg, pMessage->getAuthorId() );
    }
}

void MessageService::persistMessages( MessageList pMessage )
{
    mPersistanceLayer->transaction();

    for ( const auto &currentMessage : pMessage ) {
        if ( !currentMessage.isNull() ) {
            persistMessage( currentMessage );
        }
    };

    mPersistanceLayer->commit();
}

QSharedPointer<RocketChatMessage> MessageService::parseMessage( const QJsonObject &pMessageData, bool linkify )
{
    ChatMessage message( nullptr );
    QString userId = mServer->getUserId();
    bool blocked = false;

    if ( Q_LIKELY( pMessageData.contains( QStringLiteral( "msg" ) ) ) ) {
        //parse message String
        QString msgString = pMessageData[QStringLiteral( "msg" )].toString();
        qint16 emojiHashMsg = 0;

        if ( pMessageData.contains( QStringLiteral( "emojiHash" ) ) ) {
            emojiHashMsg = pMessageData[QStringLiteral( "emojiHash" )].toInt();
        }

        if ( linkify ) {
            msgString = Utils::removeUtf8Emojis( msgString );

            msgString = Utils::linkiFy( msgString );
            msgString = msgString.replace( QStringLiteral( "\n" ), QStringLiteral( "<br>" ) );
        }

        if ( mEmojiRepo != nullptr && !emojiHashMsg ) {
            msgString = Utils::emojiFy( msgString, mEmojiRepo );
            emojiHashMsg = 1;
            pMessageData[QStringLiteral( "emojiHash" )] = emojiHashMsg;
        }

        QString msgType;
        QString author;
        QString authorId;

        if ( Q_LIKELY( pMessageData.contains( "t" ) ) ) {
            msgType = pMessageData["t"].toString();
        }

        QJsonObject userObject = pMessageData["u"].toObject();

        if ( Q_LIKELY( userObject.contains( QStringLiteral( "username" ) ) ) ) {
            author = userObject[QStringLiteral( "username" )].toString();
        }

        // if alias given, us it instead of the username, required for bridges
        if ( Q_LIKELY( pMessageData.contains( QStringLiteral( "alias" ) ) ) ) {
            author = pMessageData[QStringLiteral( "alias" )].toString();
        }

        if ( Q_LIKELY( userObject.contains( "_id" ) ) ) {
            authorId = userObject[QStringLiteral( "_id" )].toString();
        }

        bool ownMessage = false;

        if ( authorId == userId ) {
            ownMessage = true;
        }

        QJsonObject timestampObject = pMessageData[QStringLiteral( "ts" )].toObject();
        qint64 timestamp = 0;

        if ( Q_LIKELY( timestampObject.contains( QStringLiteral( "$date" ) ) ) ) {
            timestamp = timestampObject[QStringLiteral( "$date" )].toDouble();

        }

        QDateTime date = QDateTime::fromMSecsSinceEpoch( timestamp );

        QString formattedDate = date.toString( QStringLiteral( "dd.MM.yyyy" ) );
        QString formattedTime = date.toString( QStringLiteral( "hh:mm" ) );
        auto attachments = processAttachments( pMessageData );
        message = ChatMessage( new RocketChatMessage( pMessageData ) );
        message->setAttachments( attachments );
        message->setOwnMessage( ownMessage );
        message->setAuthorId( authorId );

        if ( attachments.size() > 0 ) {
            auto firstAttachment  = attachments.first();

            if ( !firstAttachment.isNull() ) {
                msgString = firstAttachment->getTitle();
            }
        }

        message->setAuthor( author );
        message->setFormattedDate( formattedDate );
        message->setFormattedTime( formattedTime );
        message->setMessageString( msgString );
        message->setMessageType( msgType );

        if ( mBlockedUsers.contains( author ) ) {
            blocked = true;
        }
    }

    if ( Q_UNLIKELY( message.isNull() ) ) {
        qDebug() << "invalid messag";
    }

    if ( !blocked ) {
        return message;
    } else {
        return nullptr;
    }
}

void MessageService::addUserToBlockList( const QString &pId )
{
    mBlockedUsers.insert( pId );
    QMetaObject::invokeMethod( mMessagesModel, "addBlockedUser", Q_ARG( QString, pId ) );

}

void MessageService::addChannelToBlockList( const QString &pId )
{
    QString roomId = pId;
    QString userId = mServer->getUserId();
    QString otherUserId = roomId.replace( userId, "" );
    addUserToBlockList( otherUserId );
}

void MessageService::loadHistoryFromServer( QSharedPointer<LoadHistoryRequestContainer> pContainer )
{
    if ( !pContainer.isNull() ) {
        auto requests = pContainer->getRequests();
        pContainer->mAlredayReceived = 0;

        //TODO: fix memory leak risk
        QSharedPointer<QSet<QString>> duplicateCheck( new QSet<QString> );

        if ( requests.length() ) {
            QMultiMap<QString, ChatMessage> *list = new QMultiMap<QString, ChatMessage>;

            //loop over all channels
            for ( const auto &currentChannelRequest : requests ) {
                if ( !currentChannelRequest.isNull() ) {
                    QString roomId = currentChannelRequest->getChannelIds().first();
                    QJsonObject startObj ;
                    startObj[QStringLiteral( "$date" )] = currentChannelRequest->getStart();
                    QJsonObject endObj ;
                    endObj[QStringLiteral( "$date" )] = currentChannelRequest->getEnd();
                    QJsonArray params;

                    int limit = currentChannelRequest->getLimit();

                    if ( currentChannelRequest->getEnd() < 0 ) {
                        if ( currentChannelRequest->getStart() < 0 ) {
                            params = {roomId, QJsonValue::Null, limit, QJsonValue::Null};
                        } else {
                            params = {roomId, QJsonValue::Null, limit, startObj};
                        }
                    } else {

                        params = {roomId, endObj, limit, startObj};
                    }

                    QSharedPointer<DDPMethodRequest> request( new DDPMethodRequest( QStringLiteral( "loadHistory" ), params ) );
                    std::function<void ( QJsonObject, MeteorDDP * )> ddpSuccess = [ = ]( QJsonObject pResponse, MeteorDDP * pDdp ) {
                        Q_UNUSED( pDdp );

                        if ( Q_LIKELY( pResponse.contains( QStringLiteral( "result" ) ) ) ) {
                            QJsonObject result = pResponse[QStringLiteral( "result" )].toObject();

                            if ( Q_LIKELY( result.contains( QStringLiteral( "messages" ) ) ) ) {
                                QJsonArray messages  = result[QStringLiteral( "messages" )].toArray();

                                for ( auto currentMessage : messages ) {
                                    RocketChatMessage message( currentMessage.toObject() );
                                    bool newMessage = true;

                                    if ( mServer->getChannels()->contains( message.getRoomId() ) ) {
                                        auto channel = mServer->getChannels()->get( message.getRoomId() );
                                        newMessage = !channel->getMessageRepo()->contains( message.getId() );
                                    }

                                    auto messageObject = parseMessage( currentMessage.toObject(), true );

                                    if ( !messageObject.isNull() && !duplicateCheck->contains( messageObject->getId() ) && newMessage ) {
                                        list->insert( messageObject->getRoomId(), messageObject );
                                        duplicateCheck->insert( messageObject->getId() );
                                    }
                                }
                            }
                        }

                        pContainer->mAlredayReceived++;

                        if ( pContainer->mAlredayReceived == pContainer->mRequests.count() ) {
                            auto containerSuccess = pContainer->getSuccess();

                            if ( containerSuccess ) {
                                containerSuccess( list );
                            }
                        }
                    };

                    request->setSuccess( ddpSuccess );

                    mServer->sendDdprequest( request );
                }
            }
        }
    }
}

void MessageService::loadHistoryFromServer( QSharedPointer<LoadHistoryServiceRequest> pRequest )
{
    if ( !pRequest.isNull() ) {
        QList<QSharedPointer<LoadHistoryServiceRequest>> list;
        list.append( pRequest );
        QSharedPointer<LoadHistoryRequestContainer> container( new LoadHistoryRequestContainer( list, pRequest->getSuccess() ) );
        container->setSuccess( pRequest->getSuccess() );
        loadHistoryFromServer( container );
    }
}

QMultiMap<QString, ChatMessage> *MessageService::loadHistoryFromDb( QSharedPointer<LoadHistoryServiceRequest> pRequest )
{
    QMultiMap<QString, ChatMessage> *channelMap  = new QMultiMap<QString, ChatMessage>;

    if ( !pRequest.isNull() ) {
        QStringList channelIds = pRequest->getChannelIds();
        double start = pRequest->getStart();
        int limit = pRequest->getLimit();
        QList<QJsonObject> messages;
        messages.reserve( 100 );

        for ( const auto &channelId : channelIds ) {
            QList<QJsonObject> result = mPersistanceLayer->getMessagesByRid( channelId, static_cast<qint64>( start ), static_cast< int >( limit ) );
            messages.append( result );
        }

        for ( const auto &current : messages ) {
            auto parsedMessage = parseMessage( current, false );

            if ( Q_LIKELY( !parsedMessage.isNull() ) ) {
                channelMap->insert( parsedMessage->getRoomId(), parsedMessage );
            }
        }
    }

    return channelMap;
}

void MessageService::loadHistoryFromAutoSource( QSharedPointer<LoadHistoryServiceRequest> pRequest )
{
    int limit = pRequest->getLimit();
    auto channelMap = loadHistoryFromDb( pRequest );
    QList<QSharedPointer<LoadHistoryServiceRequest>> requestList;
    requestList.reserve( 50 );
    auto serverSuccess = [ = ]( QMultiMap<QString, QSharedPointer<RocketChatMessage>> *messages ) {
        if ( messages != nullptr ) {
            auto success = pRequest->getSuccess();
            success( messages );
        }
    };

    for ( const auto &key :  pRequest->getChannelIds() ) {
        if ( Q_LIKELY( channelMap->contains( key ) ) ) {
            MessageList currentList = channelMap->values( key );

            if ( currentList.count() < limit ) {
                qSort( currentList );
                double timestamp;

                if ( !currentList.empty() ) {
                    auto latestMessage = currentList.last();
                    timestamp = latestMessage->getTimestamp();
                } else {
                    timestamp = QDateTime::currentDateTime().toMSecsSinceEpoch();
                }

                QSharedPointer<LoadHistoryServiceRequest> newrequest( new LoadHistoryServiceRequest( key ) );
                newrequest->setStart( timestamp );
                newrequest->setLimit( limit - currentList.count() );

                if ( pRequest->getEnd() > -1 ) {
                    newrequest->setEnd( pRequest->getEnd() );
                }

                requestList.append( newrequest );
            }
        } else {
            QSharedPointer<LoadHistoryServiceRequest> newrequest( new LoadHistoryServiceRequest( key ) );
            double timestamp = QDateTime::currentDateTime().toMSecsSinceEpoch();

            if ( pRequest->getStart() > -1 ) {
                timestamp = pRequest->getStart();
            }

            newrequest->setStart( timestamp );

            if ( pRequest->getEnd() > - 1 ) {
                newrequest->setEnd( pRequest->getEnd() );
            }

            newrequest->setLimit( limit );
            requestList.append( newrequest );
        }

        //in all cases load the last three days
        if ( !pRequest->getNoLastThreeDays() ) {
            double end = QDateTime::currentMSecsSinceEpoch();
            double start = end - ( 86400 * 3 * 1000 );
            QSharedPointer<LoadHistoryServiceRequest> lastThreeDays( new LoadHistoryServiceRequest( key ) );
            pRequest->setStart( start );
            pRequest->setStart( start );
            requestList.append( lastThreeDays );
        }
    }

    auto success = pRequest->getSuccess();
    success( channelMap );
    qDebug() << "loaded from database now loading from sever with " << requestList.size() << "requests";
    QSharedPointer<LoadHistoryRequestContainer> requests( new LoadHistoryRequestContainer( requestList, serverSuccess ) );
    loadHistoryFromServer( requests );
}

QList<QSharedPointer<RocketChatAttachment>> MessageService::processAttachments( QJsonObject pMessageObj )
{

    QList<QSharedPointer<RocketChatAttachment> > attachmentsList;

    if ( Q_LIKELY( pMessageObj.contains( QStringLiteral( "file" ) ) ) ) {
        QJsonObject fileObj = pMessageObj[QStringLiteral( "file" )].toObject();

        if ( Q_LIKELY( fileObj.contains( QStringLiteral( "_id" ) ) ) ) {
            QJsonArray attachments = pMessageObj[QStringLiteral( "attachments" )].toArray();

            if ( Q_LIKELY( attachments.size() ) ) {
                QString type;
                QString url;
                QString title;
                int width = 100;
                int height = 100;
                QJsonObject attachment = attachments[0].toObject();

                if ( Q_LIKELY( attachment.contains( QStringLiteral( "image_url" ) ) ) ) {
                    type = QStringLiteral( "image" );
                    url = attachment[QStringLiteral( "image_url" )].toString();

                    if ( attachment.contains( QStringLiteral( "image_dimensions" ) ) ) {
                        QJsonObject imageDimensions = attachment[QStringLiteral( "image_dimensions" )].toObject();

                        if ( imageDimensions.contains( QStringLiteral( "height" ) ) && imageDimensions.contains( QStringLiteral( "width" ) ) ) {
                            height = imageDimensions[QStringLiteral( "height" )].toDouble();
                            width = imageDimensions[QStringLiteral( "width" )].toDouble();
                        }
                    }
                } else if ( attachment.contains( QStringLiteral( "video_url" ) ) ) {
                    type = QStringLiteral( "video" );
                    url = attachment[QStringLiteral( "video_url" )].toString();
                } else if ( attachment.contains( QStringLiteral( "audio_url" ) ) ) {
                    type = QStringLiteral( "audio" );
                    url = attachment[QStringLiteral( "audio_url" )].toString();
                } else if ( attachment.contains( QStringLiteral( "title_link" ) ) ) {
                    type = QStringLiteral( "file" );
                    url = attachment[QStringLiteral( "title_link" )].toString();
                }

                if ( Q_LIKELY( attachment.contains( QStringLiteral( "title" ) ) ) ) {
                    title = attachment[QStringLiteral( "title" )].toString();
                }

                QSharedPointer<RocketChatAttachment> attachmentObject( new RocketChatAttachment( url, type, title ) );
                attachmentObject->setHeight( height );
                attachmentObject->setWidth( width );

                attachmentsList.append( attachmentObject );
            }
        }
    }

    return attachmentsList;
}
